(* Copyright (c) ENS de Lyon and Inria. All rights reserved. *)

Require Import Reals  Psatz.
From Flocq Require Import Core Relative Sterbenz Operations Mult_error  Plus_error.
Require Import mathcomp.ssreflect.ssreflect.
Require Import Bayleyaux.
From Flocq Require Import Pff2Flocq.
Require Import F2SumFLT  DWPlus.

From Flocq Require Import Div_sqrt_error Sqrt.

From Interval Require Import Tactic.
Require Import square_root_flx.
Require Import square_root_aux.

Open Scope R_scope.

Set Implicit Arguments.

Section Main.

Notation two  := radix2.
Notation pow e := (bpow two e).


Variable p: Z.
Context { prec_gt_0_ : Prec_gt_0 p }.
Variable choice : Z -> bool.
Hypothesis choice_sym: forall x, choice x  = negb (choice (- (x + 1))).

Hypothesis pGt1: (1 < p)%Z.




Let u := pow (-p).

Section Underflow.

Variable emin emax:Z.

Hypothesis emin_max_p:emin = (1 - emax)%Z /\ (emin < - p - 1)%Z /\ (1 < p)%Z.

Hypothesis Hp5 : (5 <= p)%Z.




Variables xh xl:R.



Hypothesis xhpos: 0 < xh.
Variable k:Z.
Hypothesis k_emin: (2*k >= (emin +p))%Z.
Hypothesis xle_b: pow (2*k) <= (xh + xl).


Notation Emin:=  (emin - p + 1)%Z.
(* Hypothesis emax_p: (3*p +3 <= emax)%Z. *)
 Hypothesis emax_p: (p +3 <= emax)%Z.
(* first line *)





Notation fexp := (FLT_exp Emin   p).
Notation format_flt := (generic_format two  fexp ).
Notation RN :=(round two  fexp (Znearest choice) ).

(* Notation double_word_flt := (double_word emin  choice). *)
Notation double_word_flt xh xl :=  (((format_flt xh) /\ (format_flt xl)) /\ 
                                 ( xh = RN (xh + xl))%R).
Hypothesis DWx : double_word_flt  xh xl.

Notation sh  := (RN (sqrt xh)).
Notation ro1    := (RN (xh - (sh )^2)).
Notation ro2  := (RN (xl + (ro1))).
Notation sl  := (RN ((ro2 )/ (2*(sh)))).

Notation alpha := (alpha p emin).

(* Variable k: N. *)



Lemma no_uf_sqrt_xh:  pow (Emin) <= sqrt xh.
Proof.
case:DWx=>[[Fxh Fxl] xE].
case:(Rle_lt_dec 1 xh)=> xh1.
apply/(Rle_trans _ 1).
rewrite -pow0; apply/bpow_le; lia.
by rewrite -sqrt_1; apply/sqrt_le_1_alt.

apply/(Rle_trans _ xh).
rewrite -(Rabs_pos_eq xh); try lra.
apply/alpha_minFP=>//; lra.

apply/Rlt_le/sqrt_more; lra.
Qed.

Lemma no_uf_sh:  pow (Emin) <= sh.
Proof.
have->: pow Emin = RN (pow Emin).
rewrite round_generic //; apply/generic_format_FLT_bpow; lia.

apply/round_le/no_uf_sqrt_xh.
Qed.

Lemma normal_fp_sh: pow (emin) <= sh.
Proof.
case:(DWx)=> [[Fxh Fxl] xE].

have->: pow emin = RN (pow emin).
rewrite round_generic //; apply/generic_format_FLT_bpow; lia.

apply/round_le/Rsqr_incr_0_var; last by apply/sqrt_ge_0.
rewrite Rsqr_sqrt /Rsqr; last lra.
apply/(Rle_trans _ (pow (2 * k))).

rewrite -bpow_plus; apply/bpow_le; lia.
rewrite xE.
have ->: pow (2 * k) = RN (pow (2 * k)).
rewrite round_generic //; apply/generic_format_FLT_bpow; lia.
by apply/round_le.
Qed.




(* Lemma 1.5 *)

Lemma  ro1E_flt: ro1  = xh - (Rsqr (sh ) ).
Proof.
case:DWx=>[[Fxh Fxl]xE].
have pkE: pow k = RN (pow k).
  rewrite round_generic //.
  apply/generic_format_FLT_bpow; lia.
have p2kE: pow(2* k) = RN (pow (2*k)).
  rewrite round_generic //.
  apply/generic_format_FLT_bpow; lia.
have xhlb: pow (2 * k) <= xh.
  by rewrite p2kE xE; apply/round_le;lra.
have h: pow k <= sqrt xh.
  apply/Rsqr_incr_0_var.
    rewrite Rsqr_sqrt;try lra.
    by rewrite /Rsqr -bpow_plus Z.add_diag.
  by apply/sqrt_ge_0.
rewrite round_generic // -Rsqr_pow2 //.
apply/(@ro1_exact _ _ emax)=>//.
case:(@ex_mul_flt p emin sh k); try apply/generic_format_round.
rewrite Rabs_pos_eq.
    by rewrite pkE; apply/round_le.
  have ->: 0 = RN 0 by rewrite round_0.
  apply/round_le/sqrt_ge_0.
move=> m shE.
exists m ; exists k.
split; first by lia.
rewrite !bpow_plus pow1 shE; ring.
Qed.


(* Eq 25*) 
Lemma rel_error_sqrt x (Fx: format_flt x):   Rabs (RN (sqrt x)  - (sqrt x)) <= u* /(1+u) * (sqrt x).
Proof.
have ugt0 : 0 < u by rewrite /u; apply/upos.
have ult1: u < 1.
rewrite /u; apply/bpow_lt_1; lia.
case:(sqrt_error_N_FLT_ex two p choice pGt1 Emin _ x)=>//;try lia.
move=> eps.
move=> [h1 h2].
rewrite h2.
ring_simplify (sqrt x * (1 + eps) - sqrt x).
rewrite Rabs_mult Rabs_pos_eq; last by apply/sqrt_pos.
rewrite Rmult_comm.

apply/Rmult_le_compat_r; first by apply/sqrt_pos.
apply/(Rle_trans _ (1 - 1 / sqrt (1 + 2 * u_ro two p))) =>//.
have ->: (u_ro two p) = u.
rewrite /u_ro bpow_plus pow1 -/u; field.
set s := sqrt _.
suff:  1 - u * / (1 + u) <= /s by lra.
field_simplify  ( 1 - u * / (1 + u)); last lra.

rewrite /Rdiv Rmult_1_l.
apply/Rinv_le.
apply/sqrt_lt_R0; try lra.
apply/Rsqr_incr_0_var; try lra.
rewrite Rsqr_sqrt; try lra.
rewrite /Rsqr ; ring_simplify.
try nra.
Qed.

(* Eq 26 *)
Lemma MVT_sqrt'   a b (apos: 0< a ) (bpos: 0 < b): Rabs (sqrt a -sqrt b) <= /(2*  (Rmin (sqrt a) (sqrt b))) * (Rabs( a - b)).
Proof.
have min_pos: 0 < Rmin a b by  by apply/Rmin_glb_lt.
 have minsqrtpos:  0 < Rmin (sqrt a) (sqrt b).
  by apply/Rmin_glb_lt; apply/sqrt_lt_R0.

case:(MVT_abs (fun x =>  sqrt x) (fun  x  => / (2* sqrt x)) a b).
  move=> c hc;  apply/derivable_pt_lim_sqrt.
  by apply/(Rlt_le_trans _ (Rmin a b)); last lra.
  (* by apply/Rmin_glb_lt; move:(DWpos DWx); rewrite /x; lra. *)
move => c [h0 [h1 h2]].
have cpos: 0 < c by lra.
have srqtcpos:= (sqrt_lt_R0 _ cpos).
rewrite -Rabs_Ropp Ropp_minus_distr h0 -(Rabs_Ropp (b -a)) Ropp_minus_distr.
apply/Rmult_le_compat_r; first by apply/Rabs_pos.
rewrite !Rinv_mult_distr; try lra.
rewrite Rabs_mult Rabs_pos_eq ; try lra.
rewrite Rabs_Rinv; last lra.
apply/Rmult_le_compat_l; try lra.
apply/Rinv_le=>//.
rewrite Rabs_pos_eq; last lra.

case/Stdlib.Rmin_Rle: h1=>h1.


apply/(Rle_trans _ (sqrt a)).
apply/Rmin_l.
by apply/sqrt_le_1_alt.
by apply/(Rle_trans _ (sqrt b)); try apply /Rmin_r; apply/sqrt_le_1_alt.
Qed.

Notation G1 := ((2 * (sqrt (1 + u)) + 1)/(2 * (1 + u))).

Fact G1_ub : G1 < 3/2.
Proof.
have ugt0 : 0 < u by rewrite /u; apply/upos.
have ult1: u < 1 by rewrite /u; apply/bpow_lt_1; lia.
apply/(Rlt_trans _ ((2 * (1 + u/2)  + 1)/(2 * (1 + u)))).
  rewrite /Rdiv; apply/Rmult_lt_compat_r.
    by  apply/Rinv_0_lt_compat; lra.
  apply/Rplus_lt_compat_r/Rmult_lt_compat_l; try lra.
apply/Rsqr_incrst_0; try lra; try apply/sqrt_ge_0.
rewrite Rsqr_sqrt /Rsqr; try lra.
  by ring_simplify; apply/Rplus_lt_compat_r; nra.
field_simplify (2 * (1 + u / 2) + 1).
apply/(Rmult_lt_reg_r (2 * (1 + u))); first lra.
rewrite /Rdiv Rmult_assoc Rinv_l; lra.
Qed.
Notation  x := (xh + xl).

Lemma L26_1: ( x <= xh) -> Rabs (sh - sqrt x) <= (sqrt x)* u *G1.
Proof.
move=> xl_neg.
have ugt0 : 0 < u by rewrite /u; apply/upos.
have ult1: u < 1 by rewrite /u; apply/bpow_lt_1; lia.
have iult1: 0 < /(1+u) by apply/Rinv_0_lt_compat; lra.
have xpos: 0 < x  by apply/(Dwpos _ DWx).
have sqrtxpos: 0 < sqrt x by apply/sqrt_lt_R0.
case:(DWx)=> [[Fxh Fxl]xE].


have->:  (sh - sqrt x) =  (sh - sqrt xh) +  (sqrt xh - sqrt x) by ring.
apply/(Rle_trans _ _  _  (Rabs_triang _ _)).

apply/(Rle_trans _ (u/(1+u) * (sqrt (x * (1 + u))) + /(2* (sqrt x)) * (u /(1+u)) * x)); last first.
  apply/Req_le.
  rewrite sqrt_mult_alt; last lra.
  have ->:   u / (1 + u) * (sqrt x * sqrt (1 + u)) = sqrt x * u * (/(1+u) * sqrt (1+u)).
    field;lra.
  have->: / (2 * sqrt x) * (u / (1 + u)) * x = sqrt x * u * /(2 * (1+u)) .
    rewrite -{2}(Rsqr_sqrt x); last lra.
    rewrite /Rsqr; field; lra.
  field; lra.





apply/Rplus_le_compat.
  apply/(Rle_trans _  (u* /(1+u) * (sqrt xh))).
    by apply/rel_error_sqrt.
  rewrite /Rdiv; apply/Rmult_le_compat_l; first  nra.
  apply/sqrt_le_1_alt.


suff : Rabs (RN x - x) <= /2 * pow (-p +1) * Rabs x.
rewrite -{1}xE !Rabs_pos_eq; try lra.
rewrite bpow_plus -/u pow1;lra.
case:  (Rle_lt_dec    x  (pow (emin + 1)))=> h.


rewrite round_generic.
rewrite Rabs_0; last ring.
rewrite bpow_plus -/u pow1 Rabs_pos_eq;  nra.
apply/Hauser=>//.
by rewrite Rabs_pos_eq ; lra.


apply:relative_error_N_FLT; try lia.
 rewrite Rabs_pos_eq ; try lra.
apply/(Rle_trans _ (pow (emin + 1))); try lra.
apply/bpow_le; lia.

case:  (Rle_lt_dec    x  (pow (emin + 1)))=> h.
rewrite {1}xE.

rewrite round_generic=>//.
rewrite Rabs_0; last ring.
suff:  0<= (/ (2 * sqrt x) * (u / (1 + u))).
nra.
rewrite /Rdiv.
apply/Rmult_le_pos; try nra.
apply/Rlt_le/Rinv_0_lt_compat; lra.
apply/Hauser=>//.
by rewrite Rabs_pos_eq ; lra.
have {2}-> : sqrt x = (Rmin (sqrt xh) (sqrt x)).
rewrite Rmin_right //.
  apply/sqrt_le_1_alt=>//.
apply/(Rle_trans _  (/(2*  (Rmin (sqrt xh) (sqrt x))) * (Rabs( xh - x)))).


apply : MVT_sqrt'; try lra.
rewrite Rmult_assoc; apply/Rmult_le_compat_l.
apply/Rlt_le/Rinv_0_lt_compat.
rewrite Rmin_right; try lra.
  apply/sqrt_le_1_alt=>//.
rewrite {1}xE.

have: Rabs ((round two (FLX_exp p) (Znearest choice) x - x) / x) <=
  pow (- p) / (1 + pow (- p)).
apply/rel_error_u =>//; lra.
rewrite -/u Rabs_mult Rabs_Rinv ?( Rabs_pos_eq x); try lra.
move/(Rmult_le_compat_r x).
rewrite Rmult_assoc Rinv_l ?Rmult_1_r; try lra.


rewrite round_FLT_FLX.
apply; lra.
rewrite Rabs_pos_eq; try lra.
apply/(Rle_trans _ (pow (emin + 1))); try lra.
apply/bpow_le; lia.
Qed.



Notation G2 := (/(1+u) * (1 + /(2* sqrt (1-u)))).

Fact G2E : G2 = 3/2 + (1 - (1+3*u)*sqrt (1-u) )/( 2 * (1+u)* (sqrt (1-u))).
Proof.
field.
have ugt0 : 0 < u by rewrite /u; apply/upos.
have ult1: u < 1 by rewrite /u; apply/bpow_lt_1; lia.
have : 0 <sqrt (1-u) by apply/ sqrt_lt_R0; lra.
lra.
Qed.


Fact G2_ub : G2 < 3/2.
Proof.
have ugt0 : 0 < u by rewrite /u; apply/upos.
have ult1: u < 1 by rewrite /u; apply/bpow_lt_1; lia.
rewrite G2E.
rewrite -[X in _ < X]Rplus_0_r; apply/Rplus_lt_compat_l.
have h0:  0 < 2 * (1 + u) * sqrt (1 - u).
  suff : 0 < sqrt (1 - u) by nra.
  by apply/ sqrt_lt_R0; lra.
apply/(Rmult_lt_reg_r (2 * (1 + u) * sqrt (1 - u)))=>//.
rewrite /Rdiv Rmult_assoc Rinv_l ?Rmult_0_l ?Rmult_1_r ; try lra.
suff :1 <  (1 + 3 * u) * sqrt (1 - u)  by lra.
apply/Rsqr_incrst_0; try nra.
rewrite Rsqr_1 Rsqr_mult Rsqr_sqrt /Rsqr; try lra.
ring_simplify.
suff: 0 < -9 * u ^ 2 + 3 * u  + 5  by nra.
apply/(Rlt_trans _ (2 *u + 5)); try lra.

 suff:  2 * u  < -9 * u ^ 2 + 3 * u by lra.
have->: -9 * u ^ 2 + 3 * u  = (3 - 9*u ) *u by ring.
apply/Rmult_lt_compat_r; try lra.
suff: u < /16 by lra.
have ->: /16 = pow (-4) by [].
rewrite /u; apply/bpow_lt; lia.
Qed.

Lemma L26_2: (  xh < x ) -> Rabs (sh - sqrt x) <= (sqrt x)* u *G2.
Proof.
move=> xl_pos.
have ugt0 : 0 < u by rewrite /u; apply/upos.
have ult1: u < 1 by rewrite /u; apply/bpow_lt_1; lia.
have iult1: 0 < /(1+u) by apply/Rinv_0_lt_compat; lra.
have xpos: 0 < x  by apply/(Dwpos _ DWx).
have sqrtxpos: 0 < sqrt x by apply/sqrt_lt_R0.
case:(DWx)=> [[Fxh Fxl]xE].

have->:  (sh - sqrt x) =  (sh - sqrt xh) +  (sqrt xh - sqrt x) by ring.
apply/(Rle_trans _ _  _  (Rabs_triang _ _)).


pose G3:= (u/(1+u) * (sqrt x) + /(2* (sqrt( x* (1 -u)))) * (u /(1+u)) * x).



apply/(Rle_trans _  G3); last first.

  apply/Req_le.
  rewrite /G3 sqrt_mult_alt; last lra.
  have ->:   u / (1 + u) * sqrt x  = sqrt x*u * /(1+u) by field; lra.
  have->:   / (2 * (sqrt x * sqrt (1 - u))) * (u / (1 + u)) * x =
   sqrt x * u * ((/(2* sqrt (1-u)))* /(1+u)).
set A := (sqrt (1-u)).
rewrite -{2}(Rsqr_sqrt x) /Rsqr; try lra.
field.
suff : 0 < A by lra.
apply/sqrt_lt_R0; lra.
lra.




apply/Rplus_le_compat.
  apply/(Rle_trans _  (u* /(1+u) * (sqrt xh))).
    by apply/rel_error_sqrt.
  rewrite /Rdiv; apply/Rmult_le_compat_l; first  nra.
  apply/sqrt_le_1_alt; lra.
case:  (Rle_lt_dec    x  (pow (emin + 1)))=> h.
  rewrite {1}xE.
  rewrite round_generic=>//.
    rewrite Rabs_0; last ring.
have h0: 0 < sqrt (x * (1-u)).
  apply/sqrt_lt_R0; try nra.
have h1:  0<= / (2 * sqrt (x * (1 - u))).
apply/Rlt_le; rewrite Rinv_mult_distr; try lra.
suff : 0 < / sqrt (x * (1 - u)) by lra.

apply/Rinv_0_lt_compat/sqrt_lt_R0; nra.
rewrite /Rdiv;
apply/Rmult_le_pos; try lra.
apply/Rmult_le_pos; try lra.
 nra.
apply/Hauser=>//.
by rewrite Rabs_pos_eq ; lra.

apply/(Rle_trans _  (/(2*  (Rmin (sqrt xh) (sqrt x))) * (Rabs( xh - x)))).

apply : MVT_sqrt'; try lra.
rewrite Rmult_assoc.
rewrite Rmin_left; last by   apply/sqrt_le_1_alt; lra.

have h1:  Rabs (xh - x) <= u / (1 + u) * x.
  rewrite {1}xE.
  have: Rabs ((round two (FLX_exp p) (Znearest choice) x - x) / x) <=
  pow (- p) / (1 + pow (- p)).
    apply/rel_error_u =>//; lra.
  rewrite -/u Rabs_mult Rabs_Rinv ?( Rabs_pos_eq x); try lra.
  move/(Rmult_le_compat_r x).
  rewrite Rmult_assoc Rinv_l ?Rmult_1_r; try lra.
  rewrite round_FLT_FLX.
   apply; lra.
  rewrite Rabs_pos_eq; try lra.
  apply/(Rle_trans _ (pow (emin + 1))); try lra.
  apply/bpow_le; lia.



apply/Rmult_le_compat =>//; try apply/Rabs_pos.
apply/Rlt_le/Rinv_0_lt_compat.
suff : 0 <sqrt xh by lra.
by apply/sqrt_lt_R0.


apply/Rinv_le.
suff : 0 < sqrt (x * (1 - u))by lra.
 apply/sqrt_lt_R0; nra.
apply/Rmult_le_compat_l; try lra.
apply/sqrt_le_1_alt.


suff : - (xh - x) <= x*u by lra.
rewrite -(Rabs_pos_eq (- _)); try lra.
rewrite Rabs_Ropp.
apply/(Rle_trans _ (u / (1 + u) * x))=>//.
rewrite Rmult_comm; apply/Rmult_le_compat_l; try lra.
rewrite /Rdiv.
apply/(Rmult_le_reg_r (1+u)); try lra.
rewrite Rmult_assoc Rinv_l ?Rmult_1_r; try nra.
Qed.


Lemma L26_3: Rabs (sh - sqrt x) <= 3/2*u*sqrt x.
Proof.
have ugt0 : 0 < u by rewrite /u; apply/upos.
have sqrt_pos: 0 <= sqrt x by apply/sqrt_pos.
case :(Rle_lt_dec x xh)=>xxh.
  apply/(Rle_trans _ ((sqrt x * u) * G1)).
    by apply/L26_1.
  suff: sqrt x * u * G1 <= sqrt x * u * (3 / 2) by lra.
  apply/Rmult_le_compat_l; try  nra.
  by apply/Rlt_le/G1_ub.
  apply/(Rle_trans _ ((sqrt x * u) * G2)).
    by apply/L26_2.
  suff: sqrt x * u * G2 <= sqrt x * u * (3 / 2) by lra.
  apply/Rmult_le_compat_l; try  nra.
  by apply/Rlt_le/G2_ub.
Qed.

Lemma L26_4 : Rsqr (sqrt x - sh) <= 9/4*Rsqr u* x.
Proof.
have xpos: 0 < x by apply/(Dwpos _ DWx).
have ugt0 : 0 < u by rewrite /u; apply/upos.
have sqrt_pos: 0 <= sqrt x by apply/sqrt_pos.
have ->:  9 / 4 * u² * x = Rsqr ((3/2) * u * (sqrt x)).
  rewrite Rsqr_mult Rsqr_sqrt /Rsqr; lra.
apply/Rsqr_le_abs_1.
rewrite [X in _ <= X]Rabs_pos_eq.
  have->:  (sqrt x - sh) = - (sh -sqrt x) by ring.
  by rewrite Rabs_Ropp; apply/L26_3.
nra.
Qed.

Lemma L26_5 : Rabs ( x -Rsqr sh ) <= (3*u + 9/4*Rsqr u)* x.
Proof.

have xpos: 0 < x by apply/(Dwpos _ DWx).
have ugt0 : 0 < u by rewrite /u; apply/upos.
have sqrtx_pos: 0 <= sqrt x by apply/sqrt_pos.
have sqrtxh_pos: 0 <= sqrt xh by apply/sqrt_pos.
have->: x - Rsqr sh = - (Rsqr sh -x) by ring.
rewrite Rabs_Ropp.
apply/Rabs_le.
move/Rabs_le_inv : L26_3=>h.

have [h1 h2]: sqrt x -  (3 / 2 * u * sqrt x) <= sh <= 3 / 2 * u * sqrt x + sqrt x by lra.

split; last first.
  have : Rsqr sh <= Rsqr ((3 / 2) * u * sqrt x + sqrt x).
    apply/Rsqr_incr_1=>//.
      have->: 0 = RN 0 by rewrite round_0.
      by apply/round_le.
    nra.
  rewrite Rsqr_plus Rsqr_sqrt; last lra.
  rewrite Rsqr_mult Rsqr_sqrt; try lra.
  rewrite Rsqr_mult; try lra.
  have h3: sqrt x * sqrt x = x by apply/sqrt_sqrt; lra.
  have ->:  2 * (3 / 2 * u * sqrt x) * sqrt x = 3*u*x by   nra.
  rewrite /Rsqr; nra.
have: Rsqr (sqrt x - 3 / 2 * u * sqrt x) <= Rsqr sh.
  apply/Rsqr_incr_1=>//.
    suff: 0 <= 1 - 3/2 *u by nra.
    suff: 2*u  <= 1 by lra.
    rewrite -pow1 -pow0 /u -bpow_plus; apply/bpow_le; lia.
  have->: 0 = RN 0 by rewrite round_0.
  by apply/round_le.
have->:  (sqrt x - 3 / 2 * u * sqrt x) = (1 - 3 / 2 * u) * sqrt x by ring.
 rewrite Rsqr_mult Rsqr_sqrt; try lra.
rewrite Rsqr_plus Rsqr_1.
have ->:  (1 + (- (3 / 2 * u))² + 2 * 1 * - (3 / 2 * u)) * x = x  - 3*u*x + (- (3 / 2 * u))²*x by field.
move=>h3.
apply/(Rle_trans _ (- 3 * u * x + (- (3 / 2 * u))² * x)); try lra.

rewrite /Rsqr; nra.
Qed.


Lemma xl_ub:   Rabs xl <= u / (1 + u) * x.
Proof.
case:(DWx)=> [[Fxh Fxl]xE].
have xpos: 0 < x by apply/(Dwpos _ DWx).
have ugt0 : 0 < u by rewrite /u; apply/upos.
have ult1: u < 1 by rewrite /u; apply/bpow_lt_1; lia.
have iult1 : 0 < /(1 + u).
  by apply/Rinv_0_lt_compat; lra.

have{1}-> : xl = x - RN x.
 rewrite {1}xE; ring.
have->:  (x - RN x) = - (RN x -x) by ring.
rewrite Rabs_Ropp.

case:  (Rle_lt_dec    x  (pow (emin + 1)))=> h.

rewrite round_generic=>//.
rewrite Rabs_0; last ring.
rewrite/Rdiv.  apply/Rmult_le_pos; nra.

apply/Hauser=>//.
by rewrite Rabs_pos_eq ; lra.
  have: Rabs ((round two (FLX_exp p) (Znearest choice) x - x) / x) <=
  pow (- p) / (1 + pow (- p)).
    apply/rel_error_u =>//; try lra.
  rewrite -/u Rabs_mult Rabs_Rinv ?( Rabs_pos_eq x); try lra.
  move/(Rmult_le_compat_r x).
  rewrite Rmult_assoc Rinv_l ?Rmult_1_r; try lra.
  rewrite round_FLT_FLX.
   apply; lra.
  rewrite Rabs_pos_eq; try lra.
  apply/(Rle_trans _ (pow (emin + 1))); try lra.
  apply/bpow_le; lia.
Qed.



Lemma L26_6 : Rabs ( x -Rsqr sh ) <= 3*u * x.
Proof.
case:(DWx)=> [[Fxh Fxl]xE].
have xpos: 0 < x by apply/(Dwpos _ DWx).
have ugt0 : 0 < u by rewrite /u; apply/upos.
have ult1: u < 1 by rewrite /u; apply/bpow_lt_1; lia.
have iult1 : 0 < /(1 + u).
  by apply/Rinv_0_lt_compat; lra.
have sqrtx_pos: 0 <= sqrt x by apply/sqrt_pos.
have sqrtxh_pos: 0 <= sqrt xh by apply/sqrt_pos.
have sh_pos: 0 <= sh.
  have-> : 0 = RN 0 by rewrite round_0.
  by apply/round_le.


have->: (x - sh²) = (xh - sh²) + xl by ring.
apply :(Rle_trans _ _  _  (Rabs_triang _ _)).
have->:  (xh - sh²) = (sqrt xh -sh) * (sqrt xh + sh).
  by rewrite /Rsqr; ring_simplify; rewrite -!Rsqr_pow2 Rsqr_sqrt; lra.
pose v := u/(1+u).
have vpos: 0 <v  by rewrite /v /Rdiv; nra.




rewrite Rabs_mult.
have h1:  Rabs (sqrt xh - sh) <= v * sqrt xh.
  rewrite -Ropp_minus_distr Rabs_Ropp.
  by apply/rel_error_sqrt.
have h2: Rabs (sqrt xh + sh) <= 2* sqrt xh + v* sqrt xh.
  rewrite Rabs_pos_eq;last lra.
  suff: sh <= sqrt xh + v* sqrt xh by lra.
  by move/Rabs_le_inv: h1; lra.
apply/(Rle_trans _ (xh * (2*v + Rsqr v) + v*x)).
  apply/Rplus_le_compat.
    have->: xh * (2 * v + v²) = (v * sqrt xh) * ( 2 * sqrt xh + v * sqrt xh).
      have : 0 <= xh by lra.
      move/Rsqr_sqrt; rewrite /Rsqr; nra.
    by apply/Rmult_le_compat; try apply/Rabs_pos.
  by apply/xl_ub.
apply/(Rle_trans _ (  (1+v)* x * (2 * v + v²) + v * x)).
  apply/Rplus_le_compat_r/Rmult_le_compat_r.
    by rewrite /Rsqr ; nra.
  suff : xh -x <= v*x by lra.
  move/Rabs_le_inv: xl_ub.
  by rewrite /v; lra.
have->: (1 + v) * x * (2 * v + v²) + v * x = ((1 + v) * (2 * v + v²) + v) * x 
  by ring.
have->:  ((1 + v) * (2 * v + v²) + v)  = (3*u + 9*u^2 + 7*u^3)/(1+ u)^3.
rewrite /v /Rsqr; field;lra.
have->:  (3 * u + 9 * u ^ 2 + 7 * u ^ 3) / (1 + u) ^ 3 * x = 
          /((1+u)^3)* ( 1 + 3 * u + (7/3) * u^2) * (3 * u * x).
  by field; lra.
rewrite-[X in _ <= X]Rmult_1_l.
apply/Rmult_le_compat_r; first nra.
apply/(Rmult_le_reg_l ((1+u)^3)); first nra.
rewrite -Rmult_assoc Rinv_r; nra.
Qed.



Lemma L27 eps (heps: Rabs(sl - (x -Rsqr sh)/(2*sh) ) <= eps):
   Rabs ((sh + sl ) - sqrt x) < 5/4* (Rsqr u) * sqrt x  + eps.
Proof.
have ugt0 : 0 < u by rewrite /u; apply/upos.
have xpos: 0 < x by apply/(Dwpos _ DWx).
have ulti32: u <= /32 by change (pow (-p) <= pow (-5)); apply/bpow_le; lia.
have sqrtx_0:= (sqrt_lt_R0 _ xhpos).

have shpos: 0 < sh.
  apply/(Rlt_le_trans _ (pow emin)); try apply/bpow_gt_0.
  by apply/normal_fp_sh.
have sqrtxh_0:= (sqrt_lt_R0 _ xpos).

apply/(Rle_lt_trans _   ((9* Rsqr u * sqrt x)/(8 - 12*u) + eps)); last first.
  apply/Rplus_lt_compat_r.
  suff:  9 * u² / (8 - 12 * u) < 5 / 4 * u² by nra.
  apply/(Rmult_lt_reg_r (8 - 12 * u)); try lra.
  rewrite /Rdiv Rmult_assoc Rinv_l; try lra.
  rewrite /Rsqr; nra.
have->: (sh + sl - sqrt x) = 
        (sh + (x - Rsqr sh)/(2*sh) - sqrt x) + (sl -(x -Rsqr sh)/(2*sh)) by ring.
apply :(Rle_trans _ _  _  (Rabs_triang _ _)).


apply/(Rle_trans _   (Rabs (sh + (x - sh²) / (2 * sh) - sqrt x) + eps)).
  by apply/Rplus_le_compat_l.

have he: (sh + (x - sh²) / (2 * sh) - sqrt x) = 
                  (Rsqr (sh - sqrt x))/(2*sh).
  apply/(Rmult_eq_reg_r (2*sh)); try lra.
  field_simplify; try  lra.
  rewrite /Rsqr; ring_simplify.
  suff:sqrt x ^ 2 = x by lra.
  rewrite -Rsqr_pow2 Rsqr_sqrt; lra.
rewrite he Rabs_pos_eq; last first.
  apply/Stdlib.Rdiv_pos_compat; try lra.
  apply/Rle_0_sqr.
apply/(Rplus_le_compat_r).
apply/(Rle_trans _ ((9* Rsqr u)/8 * x/sh)).
  apply/(Rmult_le_reg_r (2*sh)); try lra.
  rewrite Rmult_assoc Rinv_l ?Rmult_1_r; try lra.
  apply/(Rle_trans _  (9/4*Rsqr u* x)).

  move: L26_4.
  by rewrite Rsqr_neg Ropp_minus_distr.
  apply/Req_le; field; try lra.
apply/(Rle_trans  _  ((9* Rsqr u)/8 * x/(sqrt x *( 1-3/2*u)))).
apply/Rmult_le_compat_l.
    rewrite /Rsqr; nra.
  apply/Rinv_le; try nra.
  by move/Rabs_le_inv: L26_3; lra.
have ->:   9 * u² / 8 * x / (sqrt x * (1 - 3 / 2 * u))=
           (9* Rsqr u * sqrt x)/(8 - 12*u).
  have {1} ->: x = Rsqr (sqrt x) by rewrite Rsqr_sqrt; lra.
  rewrite /Rsqr; field; lra.
suff:  9 * u² / (8 - 12 * u) <= 5 / 4 * u² by nra.
apply/(Rmult_le_reg_r (8 - 12 * u)); try lra.
rewrite /Rdiv Rmult_assoc Rinv_l; try lra.
rewrite /Rsqr; nra.
Qed.

(* Lemma sub_ro2: Rabs ro2 < pow emin -> ro2 = x - (Rsqr sh). *)
(* Proof. *)
(* move:(DWx)=> [[Fxh Fxl] xE]. *)
(* move=> hro2. *)
(* rewrite [LHS]round_generic. *)
(*    by rewrite ro1E_flt; ring. *)
(* apply/Hauser=>//; try apply/generic_format_round. *)
(* case :(Rle_lt_dec   (Rabs (xl + ro1)) ( pow (emin + 1)))=>//. *)
(* move/Rlt_le/(round_le two fexp (Znearest choice)). *)
(* rewrite RN_abs // round_generic. *)
(* suff :  pow emin < pow (emin + 1) by lra. *)
(* apply/bpow_lt; lia. *)
(* by apply/generic_format_FLT_bpow;  lia. *)
(* Qed. *)

(* Fact best_b: (11/4 * ( u^2)  + 3 * u^3) < 25/8 * (u ^2). *)
(* Proof. *)
(* have ugt0 : 0 < u by rewrite /u; apply/upos. *)
(* have xpos: 0 < x by apply/(Dwpos DWx). *)
(* suff : u <= /32 by nra. *)
(* by change (pow (-p) <= pow (-5)); apply/bpow_le; lia. *)
(* Qed. *)

Lemma L27' eps (heps: Rabs(sl - (x -Rsqr sh)/(2*sh) ) < eps):
   Rabs ((sh + sl ) - sqrt x) < 5/4* (Rsqr u) * sqrt x  + eps.
Proof.
have ugt0 : 0 < u by rewrite /u; apply/upos.
have xpos: 0 < x by apply/(Dwpos _ DWx).
have ulti32: u <= /32 by change (pow (-p) <= pow (-5)); apply/bpow_le; lia.
have sqrtx_0:= (sqrt_lt_R0 _ xhpos).

have shpos: 0 < sh.
  apply/(Rlt_le_trans _ (pow emin)); try apply/bpow_gt_0.
  by apply/normal_fp_sh.
have sqrtxh_0:= (sqrt_lt_R0 _ xpos).


have->: (sh + sl - sqrt x) = 
        (sh + (x - Rsqr sh)/(2*sh) - sqrt x) + (sl -(x -Rsqr sh)/(2*sh)) by ring.
apply :(Rle_lt_trans _ _  _  (Rabs_triang _ _)).
apply/(Rlt_le_trans _   (Rabs (sh + (x - sh²) / (2 * sh) - sqrt x) + eps)).
  by apply/Rplus_lt_compat_l.

have he: (sh + (x - sh²) / (2 * sh) - sqrt x) = 
                  (Rsqr (sh - sqrt x))/(2*sh).
  apply/(Rmult_eq_reg_r (2*sh)); try lra.
  field_simplify; try  lra.
  rewrite /Rsqr; ring_simplify.
  suff:sqrt x ^ 2 = x by lra.
  rewrite -Rsqr_pow2 Rsqr_sqrt; lra.
rewrite he Rabs_pos_eq; last first.
  apply/Stdlib.Rdiv_pos_compat; try lra.
  apply/Rle_0_sqr.
apply/(Rplus_le_compat_r).
apply/(Rle_trans _ ((9* Rsqr u)/8 * x/sh)).
  apply/(Rmult_le_reg_r (2*sh)); try lra.
  rewrite Rmult_assoc Rinv_l ?Rmult_1_r; try lra.
  apply/(Rle_trans _  (9/4*Rsqr u* x)).

  move: L26_4.
  by rewrite Rsqr_neg Ropp_minus_distr.
  apply/Req_le; field; try lra.
apply/(Rle_trans  _  ((9* Rsqr u)/8 * x/(sqrt x *( 1-3/2*u)))).
apply/Rmult_le_compat_l.
    rewrite /Rsqr; nra.
  apply/Rinv_le; try nra.
  by move/Rabs_le_inv: L26_3; lra.
have ->:   9 * u² / 8 * x / (sqrt x * (1 - 3 / 2 * u))=
           (9* Rsqr u * sqrt x)/(8 - 12*u).
  have {1} ->: x = Rsqr (sqrt x) by rewrite Rsqr_sqrt; lra.
  rewrite /Rsqr; field; lra.
suff:  9 * u² / (8 - 12 * u) <= 5 / 4 * u² by nra.
apply/(Rmult_le_reg_r (8 - 12 * u)); try lra.
rewrite /Rdiv Rmult_assoc Rinv_l; try lra.
rewrite /Rsqr; nra.
Qed.

Lemma sub_ro2: Rabs ro2 < pow emin -> ro2 = x - (Rsqr sh).
Proof.
move:(DWx)=> [[Fxh Fxl] xE].
move=> hro2.
rewrite [LHS]round_generic.
   by rewrite ro1E_flt; ring.
apply/Hauser=>//; try apply/generic_format_round.
case :(Rle_lt_dec   (Rabs (xl + ro1)) ( pow (emin + 1)))=>//.
move/Rlt_le/(round_le two fexp (Znearest choice)).
rewrite RN_abs // round_generic.
suff :  pow emin < pow (emin + 1) by lra.
apply/bpow_lt; lia.
by apply/generic_format_FLT_bpow;  lia.
Qed.

Fact best_b: (11/4 * ( u^2)  + 3 * u^3) < 25/8 * (u ^2).
Proof.
have ugt0 : 0 < u by rewrite /u; apply/upos.
have xpos: 0 < x by apply/(Dwpos _ DWx).
suff : u <= /32 by nra.
by change (pow (-p) <= pow (-5)); apply/bpow_le; lia.
Qed.

 Lemma L27_0  (hro2: Rabs ro2 < pow emin)(hro2i:pow emin <= Rabs (ro2 / (2 * sh))):
     Rabs (sl - ro2/(2*sh)) <= (3* Rsqr u* sqrt x)/(2 -3*u).
Proof.
have ugt0 : 0 < u by rewrite /u; apply/upos.
have shpos: 0 < sh.
  apply/(Rlt_le_trans _ (pow emin)); try apply/bpow_gt_0.
  by apply/normal_fp_sh.
have shipos: 0 < /sh by apply/Rinv_0_lt_compat.
have xpos: 0 < x by apply/(Dwpos _ DWx).
have sqrtx_pos: 0 < sqrt x by apply/sqrt_lt_R0.
apply/(Rle_trans _ ( u * Rabs (ro2/(2*sh)))).
  have->: u * Rabs (ro2 / (2 * sh)) = /2 * pow (-p + 1) * Rabs (ro2 / (2 * sh)).
    rewrite bpow_plus -/u pow1; lra.
  apply/(relative_error_N _ _ emin)=>//.
  by move=> k0 hk0; rewrite /fexp Zmax_left; lia.

rewrite sub_ro2 //.
apply/(Rle_trans _ (3/2* (Rsqr u)* x / sh)).

rewrite Rabs_mult(Rabs_pos_eq (/_)); last by apply/Rlt_le/Rinv_0_lt_compat;  lra.
rewrite Rinv_mult_distr; try lra.
apply/(Rle_trans _ (  u * (3*u*x) * (/ 2 * / sh))).
rewrite -Rmult_assoc; apply/Rmult_le_compat_r; try lra.
move: L26_6;  nra.
rewrite /Rsqr; lra.
apply/(Rle_trans _ (3/2* Rsqr u * x /(sqrt x * ( 1-3/2 *u)))).
  apply/Rmult_le_compat_l; first by rewrite /Rsqr; nra.  
  apply/Rinv_le.
apply/Rmult_lt_0_compat =>//.
suff : u < /2 by lra.
change (pow (-p) < pow (-1)); apply/bpow_lt; lia.
move/Rabs_le_inv: L26_3.
lra.
rewrite -{1}(sqrt_sqrt x); last lra.
apply/Req_le; field.
suff : u < /2  by lra.
change (pow (-p) < pow (-1)); apply/bpow_lt; lia.
Qed.

 Lemma L27_1 (hro2: Rabs ro2 < pow emin)(hro2i:pow emin <= Rabs (ro2 / (2 * sh))):
     Rabs (sl - ro2/(2*sh)) < (3/2 * Rsqr u + 3 * u ^3 ) *sqrt x.
Proof.
have ugt0 : 0 < u by rewrite /u; apply/upos.
have xpos: 0 < x by apply/(Dwpos _ DWx).
have sqrtx_pos: 0 < sqrt x by apply/sqrt_lt_R0.
have u32: u <= /32 by change (pow (-p) <= pow (-5)); apply/bpow_le; lia.

apply/(Rle_lt_trans _ ((3* Rsqr u* sqrt x)/(2 -3*u))); first by apply/L27_0.
have ->: (3 / 2 * u² + 3 * u ^ 3) * sqrt x =  3 * u² * sqrt x * (/2 + u).
  rewrite /Rsqr; field.
apply/Rmult_lt_compat_l; rewrite /Rsqr; try nra.
suff: 0< (u * u) * sqrt x by lra.
apply/Rmult_lt_0_compat; nra.
apply/(Rmult_lt_reg_r (2 -3*u)); try nra.
rewrite Rinv_l; nra.
Qed.

 Lemma L27_1' (hro2: Rabs ro2 < pow emin)(hro2i:pow emin <= Rabs (ro2 / (2 * sh))):
     Rabs (sh + sl -sqrt x) < (11/4 * Rsqr u + 3 * u ^3 ) *sqrt x.
Proof.
apply/(Rlt_le_trans _ (  5 / 4 * u² * sqrt x + (3 / 2 * u² + 3 * u ^ 3) * sqrt x)).
  apply/L27.
  rewrite -sub_ro2 //.
apply/Rlt_le/L27_1=>//.
lra.
Qed.

 Lemma L27_2  (hro2: Rabs ro2 < pow emin)(hro2i: Rabs (ro2 / (2 * sh)) < pow emin ):
     Rabs (sl - ro2/(2*sh)) <= pow (emin -p).
Proof.

case:(error_N_FLT two Emin p _ choice (ro2 / (2 * sh))); try lia.
move=> eps; case => eta [heps [heta [h0 slE]]].
rewrite slE.
ring_simplify (ro2 / (2 * sh) * (1 + eps) + eta - ro2 / (2 * sh)).


case:(Req_dec eta 0)=> eta0.
  rewrite eta0 Rplus_0_r Rabs_mult bpow_plus.
  apply/(Rle_trans _   (Rabs (ro2 / (2 * sh)) * pow (- p))).
    apply/Rmult_le_compat_l; first by apply/Rabs_pos.
    by move: heps; rewrite bpow_plus pow1; lra.
  apply/Rmult_le_compat_r=>//; try apply/bpow_ge_0; lra.
have -> : eps = 0 by nra.
rewrite Rmult_0_r Rplus_0_l.
apply/(Rle_trans _  (/ 2 * pow Emin))=>//.
rewrite !bpow_plus pow1; lra.
Qed.

(* powerRZ_Rpower: *)
(*   forall (x : R) (z : Z), 0 < x -> powerRZ x z = Rpower x (IZR z) *)

(* Rpower_sqrt: forall x : R, 0 < x -> Rpower x (/ 2) = sqrt x *)



 Lemma L27_2'  (hro2: Rabs ro2 < pow emin)(hro2i: Rabs (ro2 / (2 * sh)) < pow emin ):
       Rabs (sh + sl -sqrt x) < (7/4 * Rsqr u ) *sqrt x.
Proof.
have ugt0 : 0 < u by rewrite /u; apply/upos.
have xpos: 0 < x by apply/(Dwpos  _ DWx).
have sqrtx_pos: 0 < sqrt x by apply/sqrt_lt_R0.
have u32: u <= /32 by change (pow (-p) <= pow (-5)); apply/bpow_le; lia.


apply/(Rlt_le_trans _ (  5 / 4 * u² * sqrt x + pow(emin -p))).
apply/L27.
rewrite -sub_ro2 //.

apply/L27_2=>//.


suff:  pow (emin - p) <= /2 * u² * sqrt x by lra.
apply/(Rle_trans _ ((Rpower 2  (IZR(emin -3*p)/2))*sqrt x));   last first.
apply/Rmult_le_compat_r; try lra.
rewrite /Rsqr -powm1/u -!bpow_plus.
rewrite bpow_powerRZ powerRZ_Rpower ; last by rewrite  /=;  lra.


apply/Rle_Rpower; try lra.
suff:  IZR (emin - 3 * p) <= IZR (-1 + (- p + - p)) * 2 by lra.
rewrite -mult_IZR.
apply/IZR_le; lia.
rewrite bpow_powerRZ powerRZ_Rpower; last by rewrite  /=;  lra.
apply/(Rmult_le_reg_l  (Rpower 2 (-(IZR (emin - 3 * p) / 2)))).
apply/exp_pos.
rewrite  -Rmult_assoc -!Rpower_plus.
ring_simplify (- (IZR (emin - 3 * p) / 2) + IZR (emin - 3 * p) / 2).
rewrite Rpower_O; last lra.
have ->: IZR (emin -p ) = (IZR (2*emin - 2*p))/2.
suff : 2* IZR (emin -p ) = (IZR (2*emin - 2*p)) by lra.
rewrite -mult_IZR.
apply/IZR_eq; ring.
rewrite Rmult_1_l.
field_simplify  (- (IZR (emin - 3 * p) / 2) + IZR (2 * emin - 2 * p) / 2).
rewrite -opp_IZR -plus_IZR.
ring_simplify  (- (emin - 3 * p) + (2 * emin - 2 * p))%Z.
rewrite -Rpower_mult Rpower_sqrt; last by apply/exp_pos.
rewrite -powerRZ_Rpower -?(bpow_powerRZ two); last lra.
apply/sqrt_le_1_alt.
apply/(Rle_trans _ (pow (2 * k)))=>//.
apply/bpow_le; lia.
Qed.

 Lemma L27_3  (hro2: pow emin <= Rabs ro2)(hro2i: Rabs (ro2 / (2 * sh)) < pow emin ):
     /2 < sqrt x.
Proof.
have ugt0 : 0 < u by rewrite /u; apply/upos.
have xpos: 0 < x by apply/(Dwpos _ DWx).
have sqrtx_pos: 0 < sqrt x by apply/sqrt_lt_R0.
have u32: u <= /32 by change (pow (-p) <= pow (-5)); apply/bpow_le; lia.
case:(DWx)=>[[Fxh Fxl] xE].
have shpos: 0 < sh.
  apply/(Rlt_le_trans _ (pow emin)); try apply/bpow_gt_0.
  by apply/normal_fp_sh.
have shipos: 0 < /sh by apply/Rinv_0_lt_compat.
have shi2: /2 < sh.
  apply/(Rmult_lt_reg_r (/ sh))=>//.
  rewrite -Rinv_mult_distr; try  lra.
  rewrite Rinv_r; try lra.
  apply/(Rmult_lt_reg_l (Rabs ro2))=>//.
    apply/(Rlt_le_trans _ (pow (emin))) =>//;by apply/bpow_gt_0.
  rewrite Rmult_1_r -(Rabs_pos_eq (/ (2 * sh))).
    rewrite -Rabs_mult.
    by apply/(Rlt_le_trans _ (pow emin)); lra.
  rewrite Rinv_mult_distr;  lra.

case:(Rle_lt_dec (sqrt xh) (/2)) => hsqrtxh.
suff: sh <= /2 by lra.
have ->: /2 = RN (pow (-1)).
   rewrite round_generic //.
 apply/generic_format_FLT_bpow; lia.
  by apply/round_le.
have : (/ 2)² < (sqrt xh)² by apply/Rsqr_incrst_1; lra.
rewrite Rsqr_sqrt /Rsqr; try lra.
move=>h.
case: (Rle_lt_dec x (/4))=> hxi4.
suff: xh <= /4 by lra.

rewrite xE.
have ->: /4 = RN (pow (-2 )).
   rewrite round_generic //.
 apply/generic_format_FLT_bpow; lia.
  by apply/round_le.
have-> :/2 = sqrt (/4).
apply/Rsqr_inj; try lra; try apply/sqrt_ge_0.

rewrite Rsqr_sqrt /Rsqr; lra.
apply/sqrt_lt_1_alt; lra.
Qed.


 Lemma L27_3'  (hro2: pow emin <= Rabs ro2)(hro2i: Rabs (ro2 / (2 * sh)) < pow emin ):
       Rabs (sh + sl -sqrt x) < (9/4 * Rsqr u ) *sqrt x.
Proof.
have ugt0 : 0 < u by rewrite /u; apply/upos.
have xpos: 0 < x by apply/(Dwpos _ DWx).
have sqrtx_pos: 0 < sqrt x by apply/sqrt_lt_R0.
have u32: u <= /32 by change (pow (-p) <= pow (-5)); apply/bpow_le; lia.
case:(DWx)=>[[Fxh Fxl] xE].
have shpos: 0 < sh.
  apply/(Rlt_le_trans _ (pow emin)); try apply/bpow_gt_0.
  by apply/normal_fp_sh.
have: Rabs (ro2 - (x - Rsqr sh)) <= u* Rabs ro2.
  have->: (ro2 -(x -Rsqr sh))= (ro2 - (xl + ro1)).
    by rewrite {3}ro1E_flt; ring.
  have -> : u = / 2 * pow (- p + 1) by rewrite bpow_plus -/u pow1; lra.
  case: (Rle_lt_dec  (pow (Emin + p - 1)) (Rabs (xl + ro1)))=>h.
    apply/relative_error_N_FLT_round=>//.
  rewrite Rabs_0.
    move:(Rabs_pos ro2) (bpow_ge_0 two (-p +1));  nra.
  rewrite round_generic; first  ring.
  apply/Hauser =>//; try apply/generic_format_round.
  apply/(Rle_trans _ (pow (Emin + p -1))); try lra.
  apply/bpow_le; lia.
move => h.
 have h': Rabs (sl -  ro2/(2*sh))<= pow(emin - p).
  case:(error_N_FLT two Emin p _ choice (ro2 / (2 * sh))); try lia.
  move=> eps; case => eta [heps [heta [h0 slE]]].
  rewrite slE.
  ring_simplify (ro2 / (2 * sh) * (1 + eps) + eta - ro2 / (2 * sh)).
  case:(Req_dec eta 0)=> eta0.
    rewrite eta0 Rplus_0_r Rabs_mult bpow_plus.
    apply/(Rle_trans _   (Rabs (ro2 / (2 * sh)) * pow (- p))).
      apply/Rmult_le_compat_l; first by apply/Rabs_pos.
      by move: heps; rewrite bpow_plus pow1; lra.
    apply/Rmult_le_compat_r=>//; try apply/bpow_ge_0; lra.
  have -> : eps = 0 by nra.
  rewrite Rmult_0_r Rplus_0_l.
  apply/(Rle_trans _  (/ 2 * pow Emin))=>//.
  rewrite !bpow_plus pow1; lra.
apply/(Rlt_le_trans _ (  5 / 4 * u² * sqrt x + (Rsqr u)*sqrt x)); last lra.
apply/L27.
have->: sl - (x - sh²) / (2 * sh)= sl - ro2 / (2 * sh) + (ro2 - (x - Rsqr sh))/(2*sh).
  field;lra.
have sh2ipos: 0 < /(2*sh) by apply/Rinv_0_lt_compat; lra.
apply :(Rle_trans _ _  _  (Rabs_triang _ _)).
apply/(Rle_trans _ ((pow (emin - p))+ u* (Rabs ro2)/(2*sh))).
  apply/Rplus_le_compat=>//.
  rewrite Rabs_mult (Rabs_pos_eq (/_)); try lra.
  apply/Rmult_le_compat_r; lra.
apply/(Rle_trans _ (pow (emin -p +2)*sqrt x)); last first.
  apply/Rmult_le_compat_r; try lra.
  rewrite /Rsqr -bpow_plus.
  apply/bpow_le; lia.
apply/(Rle_trans _ (pow (emin -p +1))).
  apply/(Rle_trans _ (pow (emin -p ) + (pow (emin -p)))).
    apply/Rplus_le_compat_l.
    rewrite /u bpow_plus (Rmult_comm (pow emin)).
    rewrite /Rdiv Rmult_assoc.
    apply/Rmult_le_compat_l; first by apply/bpow_ge_0.
    move: hro2i; rewrite Rabs_mult (Rabs_pos_eq (/_)); lra.
  have-> :   pow (emin - p) + pow (emin - p) = 2 * (pow (emin -p)) by ring.
  rewrite -pow1 -bpow_plus; apply/bpow_le; lia.
apply/(Rmult_le_reg_l (pow (p -emin -2))).
  by apply/bpow_gt_0.
rewrite -Rmult_assoc  -!bpow_plus.
ring_simplify (p - emin - 2 + (emin - p + 2))%Z.
ring_simplify(p - emin - 2 + Emin) %Z.
rewrite powm1 pow0 Rmult_1_l.
by apply/Rlt_le/L27_3.
Qed.

(* Notation sh  := (RN (sqrt xh)). *)
(* Notation ro1    := (RN (xh - (sh  )^2)). *)
(* Notation ro2  := (RN (xl + (ro1 ))). *)
(* Notation sl  := (RN ((ro2 )/ (2*(sh)))). *)

End Underflow. 

End Main.

